**********************************************************************
*		JAZZ - RAM Disassembler Subroutine
**********************************************************************

**********************************************************************
* Name:		?DobRam
* Interface:	( #address --> $ | #address )
* Description:	If address refers to RAM variable area, then
*		drop previous stream and disassemble as is suitable
*		for the referred RAM variable. If the address refers to
*		the start of RAM then disassembles entire RAM variable
*		area.
**********************************************************************

NULLNAME ?DobRam
::
  CODE
		GOSBVL	=SAVPTR
		GOSBVL	=POP#		#address
		D1=(5)	=IRAM@
		C=DAT1	S
		C=C+C	S
		GOC	dobramGX?

dobramSX?	LC(5)	=RAMSTART	See if #70000 <= addr <= ROMPTAB
		?A<C	A
		GOYES	NOdobram
		LC(5)	=ROMPTAB
		?A>C	A
		GOYES	NOdobram
		GONC	YESdobram

dobramGX?	LC(5)	=G_RAMSTART	See if #80000 <= addr <= G_ROMPTAB
		?A<C	A
		GOYES	NOdobram
		LC(5)	=G_ROMPTAB
		?A>C	A
		GOYES	NOdobram
NOdobram
YESdobram	GOVLNG	=GPPushT/FLp
  ENDCODE

  ?SEMI		( Wasn't a RAM variable - return )
  RDROP		( Drop caller )

* Choose format table based on ROM version
  SG_ITE

ASSEMBLE

* Macro to set normal RAM variable types
RTYP	MACRO			address, type, argument
	CON(3)	$1		* Address
	CON(1)	$2		* Type
	CON(1)	($3)-1		* Field lenght
RTYP	ENDM

RTYPREP	MACRO			address, type, argument, follower count
	CON(3)	$1		* Address
	CON(1)	($2)!8		* Type with repeat
	CON(1)	($3)-1		* Field lenght
	CON(2)	$4		* Repeat count
RTYPREP	ENDM

RTYPME	MACRO			address, type, argument, repeat count
	CON(3)	$1		* Address
	CON(1)	($2)!8		* Type with repeat
	CON(1)	($3)-1		* Field lenght
	CON(2)	($4)!#80	* Self-repeat count

RTYPBUF	MACRO			address, argument, selfrepeat
	CON(3)	$1		* Address
	CON(1)	(RTYP_BUF)!8	* Buffer type
	CON(3)	$2		* Field length
	CON(2)	($3)!#80	* Self-repeat count
RTYPBUF	ENDM

* Repeat:
*	If #00		then repeat all until the end
*	If #xx!#80	then repeat itself
*	If #xx		then do following types

RTYP_SYM	EQU 0		CON(N) with possible symbol
RTYP_CON	EQU 1		CON(N) with #hex
RTYP_NIB	EQU 2		NIBHEX
RTYP_TICKS	EQU 3		TICKS (13 nibbles)
RTYP_BITS	EQU 4		CON(N) %10101..
RTYP_ROMPTAB	EQU 5		ROMPTAB
RTYP_BUF	EQU 6		Long NIBHEX


* SX Formats table
ASSEMBLE
	CON(5)	=DOHSTR
	REL(5)	SXEndFormTab

	RTYPREP	=RAMSTART,RTYP_CON,5,0

	RTYP	=IRAMMASK,RTYP_CON,5

	RTYPREP	=FAILSTK1,RTYP_CON,1,3
	RTYP	(=FAILSTK1)+1,RTYP_TICKS,13
	RTYP	(=FAILSTK1)+14,RTYP_CON,4

	RTYPREP	=FAILSTK2,RTYP_CON,1,3
	RTYP	(=FAILSTK2)+1,RTYP_TICKS,13
	RTYP	(=FAILSTK2)+14,RTYP_CON,4

	RTYPREP	=FAILSTK3,RTYP_CON,1,3
	RTYP	(=FAILSTK3)+1,RTYP_TICKS,13
	RTYP	(=FAILSTK3)+14,RTYP_CON,4

	RTYPREP	=FAILSTK4,RTYP_CON,1,3
	RTYP	(=FAILSTK4)+1,RTYP_TICKS,13
	RTYP	(=FAILSTK4)+14,RTYP_CON,4

	RTYPREP	=NEXTIRQ,RTYP_TICKS,13,2
	RTYP	=TIMECRC,RTYP_CON,4

	RTYP	=TIMEOUT,RTYP_TICKS,13
	RTYP	=TIMEOUTCLK,RTYP_CON,1
	RTYP	=LoBatTime,RTYP_CON,1
	RTYP	=StartTime,RTYP_TICKS,13
	RTYP	=FailTime,RTYP_TICKS,13
	RTYPBUF	=TESTMSG,22,2
	RTYPBUF	=SW_Image,22,1
	RTYP	=SW_ETime,RTYP_TICKS,13

	RTYPREP	=PortStat,RTYP_BITS,1,3
	RTYP	=Port1CRC,RTYP_CON,4
	RTYP	=Port2CRC,RTYP_CON,4

	RTYPREP	=COVERstate,RTYP_CON,3,3
	RTYP	=COVERsave,RTYP_CON,5
	RTYP	(=COVERsave)+5,RTYP_CON,5

	RTYPBUF	=IRAMBUFF,16,16
	RTYP	=GraphPrtHook,RTYP_NIB,11
	RTYPBUF	=uart_buffer,16,32

	RTYPREP	=uart_buf_end,RTYP_CON,2,7
	RTYP	=uart_error,RTYP_CON,1
	RTYP	=uart_buf_st,RTYP_CON,2
	RTYP	=uart_handshk,RTYP_BITS,1
	RTYP	=uart_modes,RTYP_BITS,1
	RTYP	=uart_parity,RTYP_BITS,1
	RTYP	=uart_timeout,RTYP_CON,2

	RTYP	=IOCNIB,RTYP_BITS,1

	RTYPREP	=CONFRAM,RTYP_BITS,1,9
	RTYP	(=CONFRAM)+01,RTYP_CON,5
	RTYP	(=CONFRAM)+06,RTYP_SYM,5
	RTYP	(=CONFRAM)+11,RTYP_BITS,1
	RTYP	(=CONFRAM)+12,RTYP_CON,5
	RTYP	(=CONFRAM)+17,RTYP_SYM,5
	RTYP	(=CONFRAM)+22,RTYP_BITS,1
	RTYP	(=CONFRAM)+23,RTYP_CON,5
	RTYP	(=CONFRAM)+28,RTYP_SYM,5

	RTYPREP	=CONFTAB,RTYP_CON,4,10
	RTYP	(=CONFTAB)+04,RTYP_BITS,1
	RTYP	(=CONFTAB)+05,RTYP_CON,5
	RTYP	(=CONFTAB)+10,RTYP_SYM,5
	RTYP	(=CONFTAB)+15,RTYP_BITS,1
	RTYP	(=CONFTAB)+16,RTYP_CON,5
	RTYP	(=CONFTAB)+21,RTYP_SYM,5
	RTYP	(=CONFTAB)+26,RTYP_BITS,1
	RTYP	(=CONFTAB)+27,RTYP_CON,5
	RTYP	(=CONFTAB)+32,RTYP_SYM,5

	RTYPREP	=PORT0EOS,RTYP_CON,5,3
	RTYP	=PORT1EOS,RTYP_CON,5
	RTYP	=PORT2EOS,RTYP_CON,5

	RTYPREP	=INTRAM,RTYP_NIB,16,10
	RTYP	(=INTRAM)+16,RTYP_CON,5
	RTYP	(=INTRAM)+21,RTYP_SYM,5
	RTYP	(=INTRAM)+26,RTYP_NIB,16
	RTYP	(=INTRAM)+42,RTYP_CON,3
	RTYP	(=INTRAM)+45,RTYP_NIB,16
	RTYP	(=INTRAM)+61,RTYP_NIB,16
	RTYP	(=INTRAM)+77,RTYP_NIB,16
	RTYP	(=INTRAM)+93,RTYP_SYM,5
	RTYP	(=INTRAM)+98,RTYP_SYM,5

	RTYP	=ORghost,RTYP_CON,3

	RTYPREP	=DRSTART,RTYP_CON,5,2
	RTYP	=DREND,RTYP_CON,5

	RTYP	=IREG,RTYP_CON,3
	RTYP	=SEMAPH,RTYP_CON,2
	RTYP	=IOSAVE,RTYP_BITS,1
	RTYP	=CSPEED,RTYP_CON,5
	RTYP	=INITEN,RTYP_CON,1
	RTYP	=DISABLE_KBD,RTYP_CON,1
	RTYP	=KEYSTATE,RTYP_NIB,13
	RTYPBUF	=KEYBUFFER,34,1
	RTYP	=POPPEDKEY,RTYP_CON,2
	RTYP	=DISP1CTLg,RTYP_CON,5
	RTYP	=LINENIBSg,RTYP_CON,3
	RTYP	=DISP2CTLg,RTYP_CON,5
	RTYP	=LINECOUNTg,RTYP_CON,2

	RTYPREP	=Stk0Save,RTYP_CON,5,9		Stk0Save
	RTYP	(=Stk0Save)+05,RTYP_CON,5	Stk1Save
	RTYP	(=Stk0Save)+10,RTYP_CON,5	Stk2Save
	RTYP	(=Stk0Save)+15,RTYP_CON,5	Stk3Save
	RTYP	(=Stk0Save)+20,RTYP_CON,5	Stk4Save
	RTYP	(=Stk0Save)+25,RTYP_CON,5	Stk5Save
	RTYP	(=Stk0Save)+30,RTYP_CON,5	R2[A]save
	RTYP	(=Stk0Save)+35,RTYP_CON,1	R2[S]save
	RTYP	(=Stk0Save)+36,RTYP_CON,5	R1[A]save

	RTYPREP	=SAVE_BO,RTYP_CON,1,4
	RTYP	=SAVE_LC,RTYP_CON,2
	RTYP	=SAVE_LN,RTYP_CON,3
	RTYP	=SAVE_OFFSET,RTYP_CON,5

	RTYP	=VDISP2,RTYP_CON,5
	RTYP	=ADISP,RTYP_CON,5
	RTYPME	=VDISP,RTYP_SYM,5,38
	RTYP	=RAMEND,RTYP_SYM,5
	RTYP	=AVMEM,RTYP_CON,5
	RTYP	=ERROR,RTYP_SYM,5
	RTYP	=ATTNFLG,RTYP_CON,5
	RTYP	=FIRSTPROC,RTYP_SYM,5
	RTYPME	=SysNib1,RTYP_BITS,1,20
	RTYP	=AppCount,RTYP_CON,2
	RTYP	=ITEM1LINES,RTYP_CON,1
	RTYP	=VIEWLEVEL,RTYP_CON,5
	RTYP	=DEPTHSAVE,RTYP_CON,5
	RTYP	=RNSEED,RTYP_NIB,15
	RTYP	=SAVECLK,RTYP_CON,1
	RTYP	=ALARMSDUE,RTYP_CON,1
	RTYP	=PASTDUE,RTYP_CON,1
	RTYP	=DOUSEALARM,RTYP_CON,1
	RTYP	=NOALARMSRV,RTYP_CON,1
	RTYP	=ALARM#,RTYP_CON,5
	RTYP	=PASTDUE#,RTYP_CON,5
	RTYP	=LPD_HIST,RTYP_BITS,1
	RTYP	=ANNUNCIATORS,RTYP_BITS,2
	RTYPME	=SystemFlags,RTYP_BITS,1,16
	RTYPME	=UserFlags,RTYP_BITS,1,16
	RTYP	=ELEMENT,RTYP_CON,2
	RTYP	=FIRSTCHAR,RTYP_CON,5
	RTYP	=CR_COUNT,RTYP_CON,5
	RTYP	=STACKNUM,RTYP_CON,5
	RTYP	=TOPLINE,RTYP_CON,5
	RTYP	=LASTLAST?,RTYP_CON,1
	RTYP	=HISTORYLEVEL,RTYP_CON,1
	RTYP	=LASTARGCOUNT,RTYP_CON,1
	RTYP	=LASTARGf,RTYP_CON,1
	RTYP	=LASTERROR,RTYP_SYM,5
	RTYP	=CURSOREPOSN,RTYP_CON,5
	RTYP	=CURSORROW,RTYP_CON,5
	RTYP	=CURSOROFFSET,RTYP_CON,5
	RTYPBUF	=CURSORGROB,40,1
	RTYP	=CURSORX,RTYP_CON,5
	RTYP	=CURSORY,RTYP_CON,5
	RTYP	=CURRENTMENU,RTYP_CON,2
	RTYP	=MENULEVEL,RTYP_CON,5
	RTYP	=OLDMENU,RTYP_CON,2
	RTYP	=T1COUNT,RTYP_CON,1
	RTYP	=PADCOUNT,RTYP_CON,1
	RTYPREP	=GARBSCRATCH1,RTYP_CON,5,2
	RTYP	=GARBSCRATCH2,RTYP_CON,5
	RTYP	=SAVECROSS,RTYP_NIB,10
	RTYP	=PADJSAVE1,RTYP_CON,1
	RTYPME	=PADJSAVE2,RTYP_CON,5,2
	RTYP	=KERMMODE,RTYP_CON,1
	RTYP	=DcompWidth,RTYO_CON,2
	RTYP	=FONTHEIGHT,RTYP_CON,1
	RTYP	=FONTWIDTH,RTYP_CON,1
	RTYP	=FONTCOUNT,RTYP_CON,3
	RTYP	=NODECOUNT,RTYP_CON,3
	RTYP	=OBTREELEN,RTYP_CON,3
	RTYP	=LASTOP,RTYP_CON,1
	RTYP	=LEFTTREE,RTYP_CON,3
	RTYP	=RIGHTTREE,RTYP_CON,3
	RTYP	=PARENTTREE,RTYP_CON,3
	RTYP	=PRESTACK,RTYP_NIB,7
	RTYP	=KEYLIST,RTYP_CON,5
	RTYP	=KEYLOCK,RTYP_CON,1
	RTYP	=ACCUM,RTYP_CON,1
	RTYP	=OB/EXP?,RTYP_CON,1
	RTYP	=COLWIDTH,RTYP_CON,2
	RTYP	=ENTRWISE,RTYP_CON,1
	RTYP	=PARENCOUNT,RTYP_CON,2
	RTYP	=STRETCHCOUNT,RTYP_CON,2
	RTYP	=ClkOnNib,RTYP_CON,1
	RTYP	=XmitSrvcTOut,RTYP_CON,2
	RTYP	=DelayCt,RTYP_CON









SXEndFormTab
RPL

* GX Formats table
ASSEMBLE
	CON(5)	=DOHSTR
	REL(5)	GXEndFormTab





GXEndFormTab
RPL

* Disassemble given address and formats

  GARBAGE			( No fancy methods in here )
  CODE
  ENDCODE
;
**********************************************************************








